;/************
;* AI module
;*************/

.686
.MODEL FLAT,STDCALL
OPTION CASEMAP:NONE

INCLUDE windows.inc
INCLUDE kernel32.inc
INCLUDE user32.inc

INCLUDE main.inc
INCLUDE ai.inc
INCLUDE board.inc

;/************
;* CONST
;*************/

AI_THREADSTOPWAIT = 100
AI_STATESTOPPED = 1

;/************
;* DATA
;*************/

.DATA

AiStruct AISTRUCT <>


;/************
;* CODE
;*************/

.CODE

;/*************************************
;* Resume : Appele a l'init, creation de la thread ai
;**************************************/

AiInit::
		
		push	ebx

		mov		ebx, OFFSET AiStruct
		ASSUME	ebx:PTR_AISTRUCT

		xor		eax, eax
		push	eax
		push	eax
		push	eax
		push	eax
		call	CreateEvent

		add		[ebx].EventHandle, eax
		jz		ErrorAiInit

        ;// init profondeur
        mov     [ebx].Ply, AIPLY_MEDIUM

		lea		eax, [ebx].ThreadId
		push	eax
		xor		eax, eax
		push	eax
		push	eax
		push	AiThreadEntryPoint
		push	eax
		push	eax
		call	CreateThread

		mov		[ebx].ThreadHandle, eax
		
ErrorAiInit :		

		pop		ebx

		ret

;/*************************************
;* Resume : Point d'entree de la thread
;**************************************/

AiThreadEntryPoint::

		push	ebx

		mov		ebx, OFFSET AiStruct
		ASSUME	ebx:PTR_AISTRUCT

AiWait :

		push	INFINITE
		push	[ebx].EventHandle
		call	WaitForSingleObject

		cmp		eax, WAIT_OBJECT_0
		jne		AiExit

		cmp		[ebx].State, AI_STATESTOPPED
		je		AiExit
		
        call    AiSearchMove

		push	eax
		push	eax
		push	WM_USER
		push	MainStruct.Hwnd
		call	SendMessage

		jmp		AiWait

AiExit :

		xor		eax, eax

		pop		ebx

		ret

;/*************************************
;* Find a move for the CPU
;* Returns EAX : column to play
;**************************************/

.DATA

ALIGN 4

Col     DWORD   (AIPLY_MAX+1) DUP (BOARD_NBCOL)
Note    DWORD   ((AIPLY_MAX+1)/2) DUP (MAX_NOTE, NEG_MAX_NOTE), MAX_NOTE, NEG_MAX_NOTE
BestCol DWORD   (AIPLY_MAX+1) DUP (0)

.CODE

AiSearchMove::

        push    ebp

        push    ebx
        PLAYER  TEXTEQU <ebx>

        push    esi
        PROF    TEXTEQU <esi>

        push    edi
        BESTCOL TEXTEQU <edi>

        ;// init ply
        xor     PROF, PROF

        ;// init player
        mov     PLAYER, BOARD_PLAYER2
       
NewProf :
        ;// stop searching if winning move
        call    BoardTestWin
        jge     Estimate        

        ;// estimate if max ply
        cmp     PROF, AiStruct.Ply
        je      Estimate

        ;// otherwise, find a column to play
        mov     ecx, Col[4*PROF]
        jmp     @F
FindCol :
        mov     edx, BoardHeightCopy[4*ecx]
        cmp     edx, BOARD_NBROW
        jne     ColFound
@@ :
        dec     ecx
        jns     FindCol

        ;// end if max ply
        test    PROF, PROF
        jz      Done

GoUp:

        ;// estimation dans eax
        mov     eax, Note[4*PROF]

        ;// reset estimation
        bt      PLAYER, 1
        ;// reset column
        mov     Col[4*PROF], BOARD_NBCOL
        mov     edx, 80000001h
        sub     edx, PLAYER
        mov     ecx, Col[4*PROF - 4]
        mov     Note[4*PROF], edx

        ;// change player
        xor     PLAYER, 011b
              
        ;// update note
        ;// player1 maximize, player2 minimize
        ;// if player == 1 && new_note > Note => Note = new_note
        ;// otherwise if new_note < Note => Note = new_note
        mov     ebp, Note[4*PROF - 4]
        mov     BESTCOL, BestCol[4*PROF - 4]
        xor     edx, edx
        cmp     ebp, eax
        setl    dl
        inc     dl
        xor     edx, PLAYER
        cmovnz  ebp, eax
        cmovnz  BESTCOL, ecx
        mov     Note[4*PROF - 4], ebp
        mov     BestCol[4*PROF - 4], BESTCOL

        ;// undo previous move
        call    BoardUnsetSquare

        ;// go up
        dec     PROF
        jmp     NewProf

ColFound :
        ;// player in ebx
        ;// column to play in ecx
        ;// height in edx

        ;// remember the move
        mov     Col[4*PROF], ecx

        ;// play the move
        call    BoardSetSquareAi
      
        ;// change player
        xor     PLAYER, 011b

        ;// go down in the tree
        inc     PROF
        jmp     NewProf

Estimate :
        mov     eax, BoardNote[0]
        mov     Note[4*PROF], eax
        jmp     GoUp

Done :

        ;// undo last move
        mov     Col[PROF], BOARD_NBCOL
        mov     Note[PROF], MAX_NOTE
        
        ; return best move
        mov     eax, BestCol[PROF]
    
        pop     edi
        pop     esi
        pop     ebx
        pop     ebp

        ret

;/*************************************
;* Resume : Demarre la thread de recherche
;**************************************/

AiGetMove::

		push	AiStruct.EventHandle
		call	SetEvent
	
		ret

;/*************************************
;* Resume : Termine la thread
;**************************************/

AiQuit::

		push	ebx
		push	esi
		push	edi

		mov		ebx, OFFSET AiStruct
		ASSUME	ebx:PTR_AISTRUCT

		mov		esi, [ebx].ThreadHandle
		mov		edi, [ebx].EventHandle

		mov		[ebx].State, AI_STATESTOPPED
		
		push	edi
		call	SetEvent

		push	AI_THREADSTOPWAIT
		push	esi
		call	WaitForSingleObject

		push	esi
		call	CloseHandle

		push	edi
		call	CloseHandle

		pop		edi
		pop		esi
		pop		ebx
		
		ret

END